/*

    This code is part of Gf which is free software.
    See the file COPYING for more details.

    Copyright (C) 1999  T. W. Shield, shield@aem.umn.edu

*/
#include "gfcomments.h"
#include "gfcommon.h"
#include "gfutil.h"

/**
The constructor.
*/
GfComments::GfComments()
{
#ifdef DEBUG1
	cout << " construct gfcomments \n";
#endif	
	init();
		
}

void GfComments::init()
{
	comments = NULL;
	numcomments = 0;
	xloc = NULL;
	yloc = NULL;
	loctype = NULL;
	fonts = NULL;
	colors = NULL;
	fontsize = NULL;
		
}

void GfComments::storage(int size)
{
int i;

	if(size == numcomments)
		return;

	if(size < numcomments )
	{
		for(i=size;i<numcomments;i++)
		{
			free(comments[i]);
		}
	}
		
	comments = (char **)realloc(comments,size*sizeof(char *));
	
	if( comments == NULL && size != 0)
	{
		printf("out of memory in GfComments\n");
		exit(-1);
	}
	
	if( size > numcomments )
	{
		for(i=numcomments;i<size;i++)
		{
			comments[i] = NULL;
		}
	}
	
	if(size > 0)
	{
	
		xloc = (float *)realloc(xloc,size*sizeof(float));
		if( xloc == NULL )
		{
			printf("out of memory in GfComments\n");
			exit(-1);
		}
		yloc = (float *)realloc(yloc,size*sizeof(float));
		if( yloc == NULL )
		{
			printf("out of memory in GfComments\n");
			exit(-1);
		}	
		fontsize = (float *)realloc(fontsize,size*sizeof(float));
		if( fontsize == NULL )
		{
			printf("out of memory in GfComments\n");
			exit(-1);
		}
		loctype = (int *)realloc(loctype,size*sizeof(int));
		if( loctype == NULL )
		{
			printf("out of memory in GfComments\n");
			exit(-1);
		}
		fonts = (int *)realloc(fonts,size*sizeof(int));
		if( fonts == NULL )
		{
			printf("out of memory in GfComments\n");
			exit(-1);
		}
		colors = (int *)realloc(colors,size*sizeof(int));
		if( colors == NULL )
		{
			printf("out of memory in GfComments\n");
			exit(-1);
		}
		
		if(size > numcomments)
		{
			xloc[size-1] = 0.;
			yloc[size-1] = 0.;
			loctype[size-1] = LOCATION_ROWCOL;
			fonts[size-1] = PS_DEF_FONT;
			colors[size-1] = BLACK;
			fontsize[size-1] = PS_DEF_FONTSIZE;
		}
		
	}
	else
	{
		free(xloc);
		free(yloc);
		free(fontsize);
		free(loctype);
		free(fonts);
		free(colors);
		
		init();  // make sure all pointers are NULL
	}
	
		
	numcomments = size;
		
}

/**
	Copy constructor from a reference.
*/
GfComments::GfComments(const GfComments &source) // copy constructor
{
int len,i;

	init();
	
	storage(source.numComments());
	
	for(i=0;i<numcomments;i++)
	{
		len = source.commentLen(i);
		comments[i] = (char *)malloc((len + 1)*sizeof(char));
		if(comments[i] == NULL )
		{
			printf("out of memory in gfcomments\n");
		}
		source.comment(i,comments[i]);
		
		xloc[i] = source.locationX(i);
		yloc[i] = source.locationY(i);
		loctype[i] = source.locationType(i);
		fonts[i] = source.font(i);
		colors[i] = source.color(i);
		fontsize[i] = source.fontSize(i);
	}			
}


/**
	Copy constructor from a pointer.
*/
GfComments::GfComments(const GfComments *source) // copy constructor from pointer
{
int len,i;

	init();
	
	storage(source->numComments());
	
	for(i=0;i<numcomments;i++)
	{
		len = source->commentLen(i);
		comments[i] = (char *)malloc((len + 1)*sizeof(char));
		if(comments[i] == NULL )
		{
			printf("out of memory in gfcomments\n");
		}
		source->comment(i,comments[i]);
		
		xloc[i] = source->locationX(i);
		yloc[i] = source->locationY(i);
		loctype[i] = source->locationType(i);
		fonts[i] = source->font(i);
		colors[i] = source->color(i);
		fontsize[i] = source->fontSize(i);
	}
		
}

GfComments::~GfComments()
{
	storage(0);
	
}

/**
	Deletes all the comments.
*/
void GfComments::removeAll()
{
	storage(0);
}

/**
	Prints the comments data (text and position) to os.
<pre>
	OSTREAM is either a QTextStream or ostream as set in
	gfcommon.h
	
	formating infomation is printed by format()
</pre>
*/
void GfComments::print(OSTREAM *os)
{
int i;
	for(i=0;i<numcomments;i++)
	{
		*os << comments[i]<< "\n";
		// row , col
		*os << yloc[i]<<"     "<<xloc[i]<<"\n";
	}

}

/**
	Prints the extended GF file format information to a stream.
<pre>
	can be read back in by readFormat().
</pre>
*/
void GfComments::format(OSTREAM *os)
{
int i;
	for(i=0;i<numcomments;i++)
	{
		*os << COMMENTSMAGIC <<i<<"\n";
		*os << "C\tloctype: "<<loctype[i]<<"\n";
		*os << "C\tfont: "<<fonts[i]<<"\n";
		*os << "C\tfontsize: "<<fontsize[i]<<"\n";
		*os << "C\tcolor: "<<colors[i]<<"\n";
	}

}

/**
	Reads in an extended Gf header section for the comments.
<pre>
	checks that the magic matches.
	
	returns:
		GF_OK on success
		GF_BAD on format/input error
</pre>
*/
int GfComments::readFormat(QFile *in,char *line)
{
int i;

	if(numcomments != 0)
	{
		cout << "oops in GfComments::readFormat \n";
		exit(-1);
	}
	
	i = 0;

	while(strstr(line,COMMENTSMAGIC) != NULL)
	{
	
		storage(i+1);

		in->readLine(line,MAX_LINE_LEN-1);
		if(in->atEnd())
		{
			return(GF_BAD);
		}

		if(getintval("loctype",line,&(loctype[i])) != GF_OK)
			return(GF_BAD);
			
		in->readLine(line,MAX_LINE_LEN-1);
		if(in->atEnd())
		{
			return(GF_BAD);
		}

		if(getintval("font",line,&(fonts[i])) != GF_OK)
			return(GF_BAD);
		
		in->readLine(line,MAX_LINE_LEN-1);
		if(in->atEnd())
		{
			return(GF_BAD);
		}
		
		if(getfloatval("fontsize",line,&(fontsize[i])) != GF_OK)
			return(GF_BAD);
		
		in->readLine(line,MAX_LINE_LEN-1);
		if(in->atEnd())
		{
			return(GF_BAD);
		}
		
		if(getintval("color",line,&(colors[i])) != GF_OK)
			return(GF_BAD);
		
		in->readLine(line,MAX_LINE_LEN-1);
		if(in->atEnd())
		{
			return(GF_BAD);
		}
		
		i++;
	}
	
	return(GF_OK);
	
}

OSTREAM &operator<<(OSTREAM &os, GfComments *m)
{
	m->print(&os);
	return(os);
}
OSTREAM &operator<<(OSTREAM &os, GfComments &m)
{
	m.print(&os);
	return(os);
}

/**
	The assignment operator for a deep copy.
*/
void GfComments::operator=(const GfComments &source)
{
int len,i;

	storage(source.numComments());
	
	for(i=0;i<numcomments;i++)
	{
		len = source.commentLen(i);
		comments[i] = (char *)realloc(comments[i],(len + 1)*sizeof(char));
		if(comments[i] == NULL )
		{
			printf("out of memory in gfcomments\n");
		}
		source.comment(i,comments[i]);
		
		xloc[i] = source.locationX(i);
		yloc[i] = source.locationY(i);
		loctype[i] = source.locationType(i);
		fonts[i] = source.font(i);
		colors[i] = source.color(i);
		fontsize[i] = source.fontSize(i);
	}			
}

/**
	Appends a new comment to the list of comments.
<pre>
	inputs:
		s = comment text
		x = horizontal position
		y = vertical position
		type = LOCATION_ROWCOL or LOCATION_DATA
		
		if type == LOCATION_ROWCOL then (x,y) = col,row on 80x25
		
		if type == LOCATION_DATA then (x,y) is in data coords
				and the screen location depends on the
				scales.
	returns:
	
		GF_OK -- success
		GF_BAD - bad input value.
		
	sets the color, font and font size to the default values.
</pre>
*/	
int GfComments::addComment(const char *s, float x, float y, int type)
{
int i;
	
#ifdef DEBUG1
	cout << "doing storage\n";
#endif	

	storage(numcomments+1);
	
	i = numcomments - 1;
		
	switch(type)
	{
		case LOCATION_ROWCOL:
		case LOCATION_DATA:
			loctype[i] = type;
			break;
			
		default:
			return(GF_BAD);
	}
	
#ifdef DEBUG1
	cout << "realloc for comment\n";
#endif	
	

	comments[i] = (char *)realloc(comments[i],(strlen(s) + 1)*sizeof(char));
	if(comments[i] == NULL )
	{
		printf("out of memory in gfcomments\n");
		exit(-1);
	}
	strcpy(comments[i],s);
					
	xloc[i] = x;
	yloc[i] = y;
	
	colors[i] = BLACK;
	fonts[i] = PS_DEF_FONT;
	fontsize[i] = 1.0;
	
	return(GF_OK);
	
}		

/**
	Removes the comment number i (0 to numComments()-1).
<pre>	
	returns:
		GF_OK -- success
		GF_BAD -- failure, no such comment.
</pre>
*/	
int GfComments::removeComment(int i)
{
int j;
char *ctemp;

	if( i < numcomments && i > -1)
	{
		if( i < (numcomments - 1) )
		{
			ctemp = comments[i]; // save the one to be deleted
	
			for(j=i;j<(numcomments-1);j++)
			{
				comments[j]=comments[j+1];
				xloc[j] = xloc[j+1];
				yloc[j] = yloc[j+1];
				loctype[j] = loctype[j+1];
				colors[j] = colors[j+1];
				fonts[j] = fonts[j+1];
				fontsize[j] = fontsize[j+1];
			}
			
			comments[numcomments-1] = ctemp; // put it last
		
		}
		
		storage(numcomments-1); // frees the last comment
		
		return(GF_OK);
			
	}
	else
		return(GF_BAD);

}
/**
	Changes the data for a comment (text, position and type).
<pre>	
	Can also be used to append a comment if i == numComments().
	
	returns:
		GF_OK -- success
		GF_BAD -- failure, no such comment.
</pre>
*/	
int GfComments::setComment(int i, const char *s, float x, float y, int type)
{
	if( i == numcomments)
	{
		storage(numcomments+1);
	}
	
	if( i < numcomments && i > -1)
	{
		xloc[i] = x;
		yloc[i] = y;
		switch(type)
		{
			case LOCATION_ROWCOL:
			case LOCATION_DATA:
				loctype[i] = type;
				break;
			
			default:
				loctype[i] = LOCATION_ROWCOL;
		}
		
		comments[i] = (char *)realloc(comments[i],(strlen(s) + 1)*sizeof(char));
		if(comments[i] == NULL )
		{
			printf("out of memory in gfcomments\n");
		}
		strcpy(comments[i],s);
	
		return(GF_OK);
	}
	
	return(GF_BAD);
}

/**
	Changes the data for a comment (text and position only).
<pre>	
	Can also be used to append a comment if i == numComments().
	New comments default to ROWCOL type.
	
	returns:
		GF_OK -- success
		GF_BAD -- failure, no such comment.
</pre>
*/	
int GfComments::setComment(int i,const char *s, float x, float y)
{
	if( i == numcomments)
	{
		storage(numcomments+1);
	}
	
	if( i < numcomments && i > -1)
	{
		xloc[i] = x;
		yloc[i] = y;
		
		comments[i] = (char *)realloc(comments[i],(strlen(s) + 1)*sizeof(char));
		if(comments[i] == NULL )
		{
			printf("out of memory in gfcomments\n");
		}
		strcpy(comments[i],s);
	
		return(GF_OK);
	}
	
	return(GF_BAD);
}

/**
	Sets the font for comment i (0 to numComments()-1) to f (0 to NUM_PSFONTS-1).
<pre>	
	Font choices are in gfcommon.h psfont{}.
	
	returns:
		GF_OK -- success
		GF_BAD -- failure, no such comment.
</pre>
*/	
int GfComments::setFont(int i, int f)
{
	if( i < numcomments && i > -1)
	{
		if( f > -1 && f < NUM_PSFONTS)
		{
			fonts[i] = f;
			return(GF_OK);
		}
	}
	return(GF_BAD);
}

/**
	Sets the font size for comment i (0 to numComments()-1) to s ( > 0).
<pre>	
	Default size is PS_DEF_FONTSIZE, which is currently 1.0, which is
	roughly 12 point.
	
	returns:
		GF_OK -- success
		GF_BAD -- failure, no such comment.
</pre>
*/	
int GfComments::setFontSize(int i, float s)
{
	if( i < numcomments && i > -1)
	{
		if( s > 0.)
		{
			fontsize[i] = s;
			return(GF_OK);
		}
	}
	return(GF_BAD);
}

/**
	Sets the color for comment i (0 to numComments()-1) to c (0 to NUM_PSCOLOR-1).
<pre>	
	Color choices are in gfcommon.h pscolors{}.
	
	returns:
		GF_OK -- success
		GF_BAD -- failure, no such comment.
</pre>
*/	
int GfComments::setColor(int i, int c)
{
	if( i < numcomments && i > -1)
	{
		if( c > -1 && c < NUM_PSCOLORS)
		{
			colors[i] = c;
			return(GF_OK);
		}
	}
	return(GF_BAD);
}

/**
	Returns the font index for comment i (0 to numComments()-1).
<pre>	
	Font choices are in gfcommon.h psfont{}.
	
	returns:
		>=0 -- success == font
		-1 -- failure, no such comment.
</pre>
*/	
int GfComments::font(int i) const 
{
	if( i < numcomments && i > -1)
	{

		return(fonts[i]);
	}
	else
		return(-1);
}

/**
	Returns the font size for comment i (0 to numComments()-1).
<pre>	
	Default size is PS_DEF_FONTSIZE, which is currently 1.0, which is
	roughly 12 point.
	
	returns:
		>0 -- success
		-1 -- failure, no such comment.
</pre>
*/	
float GfComments::fontSize(int i) const 
{
	if( i < numcomments && i > -1)
	{

		return(fontsize[i]);
	}
	else
		return(-1.);
}

/**
	Returns the color for comment i (0 to numComments()-1).
<pre>	
	Color choices are in gfcommon.h pscolors{}.
	
	returns:
		>=0 -- success
		-1 -- failure, no such comment.
</pre>
*/	
int GfComments::color(int i) const 
{
	if( i < numcomments && i > -1)
	{

		return(colors[i]);
	}
	else
		return(-1);
}

/**
	Returns the location Type for comment i (0 to numComments()-1).
<pre>	
	type is either LOCATION_ROWCOL or LCOATION_DATA.
	
	returns:
		>=0 -- success
		-1 -- failure, no such comment.
</pre>
*/	
int GfComments::locationType(int i) const 
{
	if( i < numcomments && i > -1)
	{

		return(loctype[i]);
	}
	else
		return(-1);
}

/**
	Returns the X postion for comment i (0 to numComments()-1).
<pre>	
	Color choices are in gfcommon.h pscolors{}.
	
	returns:
		x -- success
		0.0 -- failure, no such comment. (not really useful)
</pre>
*/	
float GfComments::locationX(int i) const 
{
	if( i < numcomments && i > -1)
	{

		return(xloc[i]);
	}
	else
		return(0.0);
}

/**
	Returns the Y postion for comment i (0 to numComments()-1).
<pre>	
	Color choices are in gfcommon.h pscolors{}.
	
	returns:
		x -- success
		0.0 -- failure, no such comment. (not really useful)
</pre>
*/	
float GfComments::locationY(int i) const 
{
	if( i < numcomments && i > -1)
	{

		return(yloc[i]);
	}
	else
		return(0.0);
}


/**
	Returns the number of comments.
<pre>	
</pre>
*/	
int GfComments::numComments() const 
{
	return(numcomments);
}

/**
	Copies comment i into s.
<pre>
	copies an empty string if comment i does not exist.
	
	There must be enough space in s to hold the comment.  See
	commentLen().
</pre>
*/	
void GfComments::comment(int i, char *s) const 
{	
	if( i < numcomments && i > -1)
	{
		if(comments[i] != NULL)
		{
			strcpy(s,comments[i]);
		}
		else
			strcpy(s,"");
	}
	else
		strcpy(s,"");
}


/**
	Returns the strlen of comment i (0 to numComments()-1).
<pre>	
	returns:
		0 on error
</pre>
*/	
int GfComments::commentLen(int i) const 
{
	if( i < numcomments && i > -1)
	{
		if(comments[i] != NULL)
			return(strlen(comments[i]));
		else
			return(0);
	}
	else
		return(0);
}

