/*

    This code is part of Gf which is free software.
    See the file COPYING for more details.

    Copyright (C) 1999  T. W. Shield, shield@aem.umn.edu
    
    
    NOTES:
    
    	need to fix read/write when using delimiters. (untested)
	
	need to test append of extended file with different profile
	
	profile problem -- should set it for read, also can glean
		from extended -- but need to handle errors. (untested)
		
	be nice to return details of error on read/append instead of 
		just GF_BAD.
	
	

*/
#include "gffile.h"
#include <qfileinfo.h>
#include "gfutil.h"


/**
	Default constructor.
<pre>
	GfFile class is the container class for all of the data and
	settings needed to produce a plot.  It also provides file i/o
	for the Gf format files.
	
	Drawing is provided by the GfDraw class which is not contained in
	GfFile.
	
	GfFile constains:
		GfCurves (0 to numCurves() of them)
		GfScale
		GfComments
		GfPlotDefs
		GfProfile
</pre>
*/
GfFile::GfFile()
{
#ifdef DEBUG1
	cout << " construct gffile \n";
#endif	
	init();
		
}

void GfFile::init()
{
	curves = NULL;
	
	numcurves = 0;
	
	file_name = "";
	in_file_comments = "";
	
	xyscale.defaults();
	
	gfcomments.removeAll();
	
	gfplotdefs.defaults();	
	
	gfprofile.setEof(EOF_AUTO);	
}

void GfFile::storage(int size)
{
int i;

	if(size == numcurves)
		return;

	if(size < numcurves )
	{
		for(i=size;i<numcurves;i++)
		{
			delete(curves[i]);
		}
	}
		
	curves = (GfCurve **)realloc(curves,size*sizeof(GfCurve *));
	if( curves == NULL && size != 0)
	{
		printf("out of memory in GfFile\n");
		exit(-1);
	}
	
	if( size > numcurves )
	{
		for(i=numcurves;i<size;i++)
		{
			curves[i] = new GfCurve();
		}
	}
	
	if(size == 0)
	{
		curves = NULL;
	}
	
	numcurves = size;
		
}
	

/**
	Copy Constructor from a reference.
*/
GfFile::GfFile(const GfFile &source) // copy constructor
{
int i;
	init(); 
	
	storage(source.numCurves());
	
	if(numcurves > 0)
	{
		for(i=0;i<numcurves;i++)
		{
			*(curves[i]) = source.curve(i);
		}
	}
	
	xyscale = source.scale();
	
	gfprofile = source.profile();
	
	gfcomments = source.comments();
	
	gfplotdefs = source.plotDefs();
	
	file_name = source.fileName();
	
	in_file_comments = source.inFileComments();
				
}

/**
	Copy Constructor from a pointer.
*/
GfFile::GfFile(const GfFile *source) // copy constructor from pointer
{
int i;
	init(); 
	
	storage(source->numCurves());
	
	if(numcurves > 0)
	{
		for(i=0;i<numcurves;i++)
		{
			*(curves[i]) = source->curve(i);
		}
	}
	
	xyscale = source->scale();
	
	gfprofile = source->profile();
	
	gfcomments = source->comments();
	
	gfplotdefs = source->plotDefs();
	
	file_name = source->fileName();
	
	in_file_comments = source->inFileComments();
		
}

GfFile::~GfFile()
{
	storage(0);
	
}

/**
	This writes the Gf extended file format.
<pre>
	
	extended format is usual gf file format with comments at
	the begining that contain all objects formatting information.
	
	The first line of comments for the extended format contains
	the GFMAGIC string (see gfcommon.h).  Regular in file comments
	can preceed the extended file format section, but cannot follow.
	
	This information is in the following order:
	
	gfprofile
	curves: 1 ... numcurves
	gfscale
	gfcomments
	gfplotdefs
	
	then follows the standard Gf file format.
	
	on append that gfprofile and curves extended info is read, but the
	scale, comments and plotdefs are discarded, as is these parts
	of the std. file section.
	
	changes from old file format:
	
		using delimiters: invec is now written unless vlen==0.  also 
		scale xmax,xmin, etc are written with delimiters.
		
	BUGS:
		under windows E format is 1.234E+000 and gives field
		width one wider than expected. -- fixed 7/23/99
	
</pre>
*/
void GfFile::print(OSTREAM *os)
{
int i,vlen,dlen,j;
char fmt[20],line[100],ticks,*xlabel,*ylabel,temp[100];

	if(! in_file_comments.isEmpty())
	{
		*os << in_file_comments;
		if(in_file_comments.right(1) != "\n")
		{
			*os<<"\n";
		}
	}
	
	*os << GFMAGIC << GFEXTENDVER << "\n"; 
	
	gfprofile.setEof(EOF_NO);
	
	*os << gfprofile;
	
	for(i=0;i<numcurves;i++)
	{
		if(curves[i]->numPoints() > 0)
		{
			*os << CURVEMAGIC <<i+1<<"\n";
			curves[i]->format(os);
		}
	}
	
	xyscale.format(os);
	gfcomments.format(os);
	
	*os << gfplotdefs;
	
	// now for old GF file format
	
	vlen = gfprofile.invecLength();
	dlen = gfprofile.dataLength();
		
	*os << numPoints() << "\n";
	
	for(i=0;i<numcurves;i++)
	{
		for(j=0;j<curves[i]->numPoints();j++)
		{
			curves[i]->sprintRow(line,j,vlen,dlen,gfprofile.lineFormat(),gfprofile.delimChar());	
			*os << line;
		}
	}
	
        if(gfprofile.lineFormat() == FMT_FIXED)
        {
		sprintf(fmt,"%%+%i.%iE",dlen,dlen-7);

		sprintf(temp,fmt,xyscale.xMin());
#ifndef UNIX			
		if(strlen(temp) > dlen) 
			winFixExpFmt(temp);
#endif
		strcpy(line,temp);

		sprintf(temp,fmt,xyscale.xMax());
#ifndef UNIX			
		if(strlen(temp) > dlen) 
			winFixExpFmt(temp);
#endif
		strcat(line,temp);
		
         	if(xyscale.xLogScale() == GF_YES) 
		{
			strcat(line,"L");
		}
		
		strcat(line,"\n");
				
		*os << line;

		sprintf(temp,fmt,xyscale.yMin());
#ifndef UNIX			
		if(strlen(temp) > dlen) 
			winFixExpFmt(temp);
#endif
		strcpy(line,temp);

		sprintf(temp,fmt,xyscale.yMax());
#ifndef UNIX			
		if(strlen(temp) > dlen) 
			winFixExpFmt(temp);
#endif
		strcat(line,temp);
		
         	if(xyscale.yLogScale() == GF_YES) 
		{
			strcat(line,"L");
		}
 
		strcat(line,"\n");
				
 		*os << line;

		if(xyscale.xLabelLen() > 0)
		{
			strmalloc(&xlabel,xyscale.xLabelLen());
			xyscale.xLabel(xlabel);
			*os <<xlabel<<"\n";
			free(xlabel);
		}
		else
			*os <<"\n";
		
		if(xyscale.yLabelLen() > 0)
		{
			strmalloc(&ylabel,xyscale.yLabelLen());
			xyscale.yLabel(ylabel);
			*os <<ylabel<<"\n";
			free(ylabel);
		}
		else
			*os<<"\n";

		switch(xyscale.ticksType())
		{
			case TT_TICKS:
				ticks='t';
			break;
			case TT_GRIDS:
				ticks='g';
			break;
			default:
				ticks='n';
			break;
		}
	
		sprintf(line,"%c",ticks);
		
		sprintf(temp,fmt,xyscale.xTicksStep());
#ifndef UNIX			
		if(strlen(temp) > dlen) 
			winFixExpFmt(temp);
#endif
		strcat(line,temp);

		sprintf(temp,fmt,xyscale.yTicksStep());
#ifndef UNIX			
		if(strlen(temp) > dlen) 
			winFixExpFmt(temp);
#endif
		strcat(line,temp);				
		
		strcat(line,"\n");
				
		*os << line;
	} // FMT_FIXED
	else
       {
        	if(xyscale.xLogScale() == GF_YES)
        	{
                	sprintf(fmt,"%%g%c%%gL\n",gfprofile.delimChar());
        	}
        	else
        	{
                	sprintf(fmt,"%%g%c%%g\n",gfprofile.delimChar());
        	}

        	sprintf(line,fmt,xyscale.xMin(),xyscale.xMax());
		*os << line;

        	if(xyscale.yLogScale() == GF_YES)
        	{
                	sprintf(fmt,"%%g%c%%gL\n",gfprofile.delimChar());
        	}
        	else
        	{
                	sprintf(fmt,"%%g%c%%g\n",gfprofile.delimChar());
        	}

        	sprintf(line,fmt,xyscale.yMin(),xyscale.yMax());
		*os << line;

	
		if(xyscale.xLabelLen() > 0)
		{
			strmalloc(&xlabel,xyscale.xLabelLen());
			xyscale.xLabel(xlabel);
			*os <<xlabel<<"\n";
			free(xlabel);
		}
		else
			*os <<"\n";
		
		if(xyscale.yLabelLen() > 0)
		{
			strmalloc(&ylabel,xyscale.yLabelLen());
			xyscale.yLabel(ylabel);
			*os <<ylabel<<"\n";
			free(ylabel);
		}
		else
			*os<<"\n";

		switch(xyscale.ticksType())
		{
			case TT_TICKS:
				ticks='t';
			break;
			case TT_GRIDS:
				ticks='g';
			break;
			default:
				ticks='n';
			break;
		}
	
                sprintf(fmt,"%%c%%g%c%%g\n",gfprofile.delimChar());
			
        	sprintf(line,fmt,ticks,xyscale.xTicksStep(),xyscale.yTicksStep());
		*os << line;
	} // FMT_DELIM
		
	*os << gfcomments;

}

OSTREAM &operator<<(OSTREAM &os, GfFile *m)
{
	m->print(&os);
	return(os);
}
OSTREAM &operator<<(OSTREAM &os, GfFile &m)
{
	m.print(&os);
	return(os);
}

/**
	The = operator for this class.
*/
void GfFile::operator=(const GfFile &source)
{
int i;
	
	storage(source.numCurves());
	
	if(numcurves > 0)
	{
		for(i=0;i<numcurves;i++)
		{
			*(curves[i]) = source.curve(i);
		}
	}
	
	xyscale = source.scale();

	gfprofile = source.profile();
	
	gfcomments = source.comments();
	
	gfplotdefs = source.plotDefs();
	
	file_name = source.fileName();
	
	in_file_comments = source.inFileComments();
}

/**
	Reads a Gf data file.	
<pre>
	Call with:
	
		filename -- filename and path of file to open.
		xcol - data column to use for x axis
		ycol - data column to use for the y axis
		append == GF_YES to append data to existing data.
		
	returns:
		GF_BAD on file open or other read failure.
		GF_OK otherwise.
</pre>
*/
int GfFile::read(char *filename,int xcol, int ycol,int append)
{
int rtn;
	QFile qf(filename);
	
	if( qf.open(IO_ReadOnly ) )
	{
		rtn=read(&qf,xcol,ycol,append);
		qf.close();
		return(rtn);
	}
	else
		return(GF_BAD);

}

/**
	Same as read(char *filename  but takes a QFile as input.
<pre>
	 called by read(char *filename, ...)
</pre>
*/
int GfFile::read(QFile *in,int xcol, int ycol,int append)
{
char line[MAX_LINE_LEN],line2[MAX_LINE_LEN],fmt[50],fmt2[50];

int first_time,ncurv,np,row,col,pts,iv,orig_numcurves,j;
int extended,ncom;
float x,y;
char ticks;
GfProfile tprofile;

	// if we are not appending, dump the current info
	// except for the profile

	if(numcurves != 0 && append != GF_YES)
	{
		storage(0);
		init();
		gfcomments.removeAll();
		xyscale.defaults();
		gfplotdefs.defaults();
	}
	
	tprofile = gfprofile; // save a copy of the profile in case read of
			// extended data fails
	
	if(numcurves == 0) append = GF_NO;
	
	first_time = 1;
	
	extended = 0;
	orig_numcurves = -1;
	
	if(append == GF_NO)
	{
		in_file_comments.resize(0);
		ncurv = -1;
	}
	else
	{
		orig_numcurves = numcurves;
		ncurv = numcurves-1;
	}
		
	pts = 0;
	np = 1;

	while( ! in->atEnd() && pts < np)
	{	
		in->readLine(line,MAX_LINE_LEN-1);
		
		if( first_time )
		{
			while(line[0] == 'c' || line[0] =='C' )
			{
#ifdef DEBUG1
			cerr<<"line:"<<line<<"\n";
#endif			
				if(strstr(line,GFMAGIC) != NULL)
				{
#ifdef DEBUG1
					cout <<"extended file format found \n";
#endif										
					extended = 1;
					
					if(read_extended(in,line,append) != GF_OK)
					{
						extended = -1; // oops, dump rest of comments
						gfprofile = tprofile;
						if( append == GF_NO )
						{
							storage(0);
							init();
							gfcomments.removeAll();
							xyscale.defaults();
							gfplotdefs.defaults();
						}
						else // if appending only add curves extended info
						{
							if(numcurves > orig_numcurves) // we added some curves via extended fmt
							{
								for(j=numcurves-1;j>=orig_numcurves;j--)
								{
									removeCurve(j);
								}
							}
						}
							
#ifdef DEBUG1
						cout << " error on extended \n";
#endif						
					}
					else
					{
						break;
					}
				}
				
				if( extended == 0)
				{
					strip_crlf(line);
					in_file_comments.append(line);
					in_file_comments.append("\n");
				}
			
				in->readLine(line,MAX_LINE_LEN-1);
				if(in->atEnd())
				{
					goto bad;
				}
			}
			
			first_time = 0;
			
			np = auto_np(line);

#ifdef DEBUG1
			cout << "np = "<<np<<"\n";
#endif			
			if(gfprofile.eof() == EOF_NO)
			{
				if(np > 0)
					continue;
				else
				{
					gfprofile.setEof(EOF_AUTO);
					return(GF_BAD);
				}
			}
		}
		
#ifdef DEBUG1
		cout << "in:"<<line;
#endif	
	
		if(parse_data(line,&iv,&x,&y,xcol,ycol) != GF_OK)
		{
			goto bad;
		}
		
#ifdef DEBUG1
		cout << "out:"<<x<<" "<<y<<"\n";
#endif		
		
		if(iv == 0 || pts == 0 ) ncurv++;
		
		if(appendPoint(ncurv,iv,x,y) != GF_OK)
			goto bad;
		
		pts++;
		if(gfprofile.eof() == EOF_YES)
		{
			np = pts + 1;
		}
		
	}
	
	if(append == GF_YES)
	{
		gfprofile = tprofile; // may have loaded appended file profile
		goto done;
	}
		
        if(in->atEnd()) 
	{
		autoScale();
		goto done;
	}

	if(gfprofile.lineFormat() == FMT_FIXED)
	{
        	sprintf(fmt,"%%%ig%%%ig",gfprofile.dataLength(),gfprofile.dataLength());

		in->readLine(line,MAX_LINE_LEN-1);
		parse_data_fixed(line,&iv,&x,&y,1,2,0);

                xyscale.setXMaxMin(y,x);

                if(strlen(line) > (unsigned int)(2*gfprofile.dataLength()+1))
                {
                        if(line[strlen(line)-2] == 'L') xyscale.setXLogScale(GF_YES);
                }

		if(in->atEnd()) goto bad;
                in->readLine(line,MAX_LINE_LEN-1);

                parse_data_fixed(line,&iv,&x,&y,1,2,0);

                xyscale.setYMaxMin(y,x);

                if(strlen(line) > (unsigned int)(2*gfprofile.dataLength()+1))
                {
                       	if(line[strlen(line)-2] == 'L') xyscale.setYLogScale(GF_YES);
                }

		if(in->atEnd()) goto bad;
                in->readLine(line,MAX_LINE_LEN-1);
                strip_crlf(line);
		xyscale.setXLabel(line);
                
		if(in->atEnd()) goto bad;
                in->readLine(line,MAX_LINE_LEN-1);
                strip_crlf(line);
		xyscale.setYLabel(line);

                strcpy(fmt2,"%c");
                strcat(fmt2,fmt);

		if(in->atEnd()) goto bad;
                in->readLine(line,MAX_LINE_LEN-1);
			
                sscanf(line,fmt2,&ticks,&x,&y);

		if(ticks == 't' || ticks == 'T')
				xyscale.setTicksType(TT_TICKS);
		if(ticks == 'g' || ticks == 'G')
				xyscale.setTicksType(TT_GRIDS);

                xyscale.setYTicksStep(y);
                xyscale.setXTicksStep(x);

	} // FMT_FIXED read scaling
	else
	{
		in->readLine(line,MAX_LINE_LEN-1);
		
		sprintf(fmt,"%%g%c%%g",gfprofile.delimChar());

		strip_crlf(line);
		
		sscanf(line,fmt,&x,&y);
                xyscale.setXMaxMin(y,x);

                if(line[strlen(line)-1] == 'L') 
		{
			xyscale.setXLogScale(GF_YES);
		}
		
		if(in->atEnd()) goto bad;
                in->readLine(line,MAX_LINE_LEN-1);

		strip_crlf(line);
		
		sscanf(line,fmt,&x,&y);
                xyscale.setYMaxMin(y,x);
		
                if(line[strlen(line)-1] == 'L') 
		{
			xyscale.setYLogScale(GF_YES);
		}

		if(in->atEnd()) goto bad;
                in->readLine(line,MAX_LINE_LEN-1);
                strip_crlf(line);
		xyscale.setXLabel(line);
                
		if(in->atEnd()) return(GF_BAD);
                in->readLine(line,MAX_LINE_LEN-1);
                strip_crlf(line);
		xyscale.setYLabel(line);

                strcpy(fmt2,"%c");
                strcat(fmt2,fmt);

		if(in->atEnd()) goto bad;
                in->readLine(line,MAX_LINE_LEN-1);
			
                sscanf(line,fmt2,&ticks,&x,&y);

		if(ticks == 't' || ticks == 'T')
				xyscale.setTicksType(TT_TICKS);
		if(ticks == 'g' || ticks == 'G')
				xyscale.setTicksType(TT_GRIDS);

                xyscale.setYTicksStep(y);
                xyscale.setXTicksStep(x);

	} // FMT_DELIM read scaling
	
	// read comments
	
	ncom = 0;
	
        while(!in->atEnd())
	{
                in->readLine(line,MAX_LINE_LEN-1);
				
		if(in->atEnd()) break;
				
		in->readLine(line2,MAX_LINE_LEN-1);

                if(line[0] == 10 ) continue;
                if(line[0] == '\0') continue;
                if(strlen(line) == 0 ) continue;

                strip_crlf(line);
                sscanf(line2,"%i   %i",&row,&col);
                gfcomments.setComment(ncom,line,col,row);
		ncom ++;					
	}	

		
done:	
	{
		QFileInfo qfi(*in);
	
		if(append == GF_NO)
			file_name=qfi.fileName();
		else
		{
			file_name+="+";
			file_name+=qfi.fileName();
		}
	
		gfplotdefs.setPlotName(file_name);
	}
	
	return(GF_OK);
	
bad:
	if(append == GF_YES)
	{
		if(numcurves > orig_numcurves) // we added some curves before error
		{
			for(j=numcurves-1;j>=orig_numcurves;j--)
			{
				removeCurve(j);
			}
		}
	}
	else
	{
		clear(); // clear if error on read
	}
	
	return(GF_BAD);
}

int GfFile::read_extended(QFile *in,char *line,int append)
{
GfCurve *tcurve;
float extended_version;

	// need to pass extended_version into readFormat routines
	// if the extended format is ever changed from 1.0

	if(getfloatval(GFMAGIC,line,&extended_version) != GF_OK)
		return(GF_BAD);
		
	if(extended_version > GFEXTENDVER) // old version of code.
	{
		cout << "File format is newer than code, please upgrade\n";
		exit(-1);
	}

	in->readLine(line,MAX_LINE_LEN-1);
	if(in->atEnd())
	{
		return(GF_BAD);
	}
	
	if(gfprofile.readFormat(in,line)!= GF_OK)
		return(GF_BAD);	
		
	while(strstr(line,CURVEMAGIC) != NULL)
	{
		tcurve = new GfCurve();
		if(tcurve->readFormat(in,line) == GF_OK)
		{
			appendCurve(*tcurve);
		}
		else
		{
			delete tcurve;
			return(GF_BAD);
		}
		delete tcurve;
	}
	
	if(append == GF_NO)
	{
	
		if(xyscale.readFormat(in,line) != GF_OK)
			return(GF_BAD);
	
		if(gfcomments.readFormat(in,line)!= GF_OK)
			return(GF_BAD);	
		
		if(gfplotdefs.readFormat(in,line)!= GF_OK)
			return(GF_BAD);	
		
	}
	
	// make sure all comments are gone
	
	while(line[0] == 'c' || line[0] == 'C')
	{
		in->readLine(line,MAX_LINE_LEN-1);
		if(in->atEnd())
		{
			return(GF_BAD);
		}				
	}
	
	return(GF_OK);		

}

void GfFile::strip_crlf(char *text)
{
/*  
	strips any trailing \r or \n  from text 
*/
int len;
	len = strlen(text);
	
	if(len < 1) return; /* empty string */
	
	len=len-1;
	
	while(text[len] == '\r' || text[len] == '\n')
	{
		text[len]='\0'; /* null it out */
		len--;
		if(len < 0) return;
	}
}

int GfFile::parse_data_fixed(char *line,int *i1, float *x, float *y,int xcol,int ycol,int vlen)
{
int dlen,ncol;
char part[50],*temp;

        ncol = chk_line_size(line);
        if((ncol < xcol)||(ncol < ycol))
                return(GF_BAD);  /* error line not long enough */
		
	dlen = gfprofile.dataLength();

        if( vlen > 0 )
        {
		strncpy(part,line,vlen);
                part[vlen]='\0';
                sscanf(part,"%i",i1);
        }
        else
        {
                *i1=1;
        }

        temp=&line[vlen+(xcol-1)*dlen];

 	strncpy(part,temp,dlen);
 	part[dlen]='\0';
        sscanf(part,"%g",x);

        temp=&line[vlen+(ycol-1)*dlen];

 	strncpy(part,temp,dlen);
	part[dlen]='\0';
        sscanf(part,"%g",y);
	
#ifdef DEBUG1
	cout <<"pdf: "<<*x<<" "<<*y<<"\n";
#endif	

        return(GF_OK); /* every thing ok */


}


int GfFile::parse_data(char *line,int *i1, float *x, float *y,int xcol,int ycol)
{
char part[50];
char *tline,*dptr;
int i,k;

      if(gfprofile.lineFormat() == FMT_FIXED)
      {
      	return(parse_data_fixed(line,i1,x,y,xcol,ycol,gfprofile.invecLength()));
      }
      else
      {
          tline=line; /* get line start */
strip2:
          if(tline[strlen(tline)-1] == gfprofile.delimChar()) /* remove trailing delims */
          {
              tline[strlen(tline)-1] = '\0';
              goto strip2;
          }
	  
	  if(gfprofile.invecLength() > 0)
	  {
	  	dptr = strchr(tline,gfprofile.delimChar()); /* find next delimiter */
		if( dptr != NULL)
		{
			*dptr = '\0';
			*i1 = atoi(tline);
			tline = dptr + 1;
		}
		else
		{
			return(GF_BAD);
		}	  	  
	  }
	  else
	  {
	            *i1 = 1;
	  }

          i = 1; /* current data column */
          k = 0; /* number of values found: (x,y) == 2 */
more:
          dptr = strchr(tline,gfprofile.delimChar()); /* find next delimiter */

          if((dptr == NULL)&&!((k == 1)&&((xcol == i)||(ycol == i))))
          {
                return (GF_BAD); /* error */
          }

          if(dptr != NULL)
          {
               *dptr = '\0';  /* make delimiter eos */
          }

          strncpy(part,tline,50);  /* pull off column up to delimiter */

          if(dptr != NULL)
          {
                tline = dptr + 1; /* move along the line */
strip:
                if(tline[0] == ' ')  /* remove extra blanks from front */
                {
                        tline = tline+1;
                        goto strip;
                }

          }

          if(xcol == i)
          {
               *x=atof(part);
               k++;
          }
          if(ycol == i)
          {
               *y=atof(part);
               k++;
          }

          i++; /* ready for next column */

          if(k < 2) goto more;/* if we haven't got all our data yet */

          return(GF_OK); /* every thing ok */

      }

}

int GfFile::auto_np(char *line)
{
char *tline;
int np;

	np = 0;
	
#ifdef DEBUG1
	cout <<"in auto np "<<line<<"\n";
#endif	
	
	switch(gfprofile.eof())
	{
	
	case EOF_AUTO:

        switch(gfprofile.lineFormat())
        {
	
		case FMT_FIXED:
			if (strlen(line) < 2*(unsigned int)gfprofile.dataLength() )  /* check for np */
           		{
                		np=atoi(line);
                		gfprofile.setEof(EOF_NO);
           		}
           		else
           		{
                 		gfprofile.setEof(EOF_YES);
           		}
	   	break;
		
		case FMT_DELIM:
        
          		tline=line;
strip:
           		if(tline[0] == ' ') /* remove leading blanks */
            		{
               	 		tline = tline + 1;
               	 		goto strip;
            		}
strip2:
            		if(tline[strlen(tline)-1] == gfprofile.delimChar()) /* remove trailing delims */
            		{
            			tline[strlen(tline)-1] = '\0';
            			goto strip2;
            		}

            		if(strchr(tline,gfprofile.delimChar())==NULL)
            		{
            			sscanf(tline,"%i", &np);
            			gfprofile.setEof(EOF_NO);
            		}
            		else
            		{
            			gfprofile.setEof(EOF_YES);
            		}
		break;
			
        }
	
	break;
	
	case EOF_YES:
		break;
		
	case EOF_NO:
		np = atoi(line);
		break;
	}
	
	return(np);
}


/**
	Returns the number of columns of data in a gf file name filename.
<pre>
	Just calls checkNumCols(QFile *qf).
		
	returns:
		-1 on error
		2 or more is number of data columns.
</pre>
*/
int GfFile::checkNumCols(char *filename)
{
int n;

	QFile qf(filename);
	
	if( qf.open(IO_ReadOnly ) )
	{
		n = checkNumCols(&qf);
		qf.close();
		return(n);
	}
	else
		return(-1);
		
}

/**
	Returns the number of columns of data in a gf file pointed to by in.
<pre>	
	returns:
		-1 on error
		2 or more is number of data columns.
</pre>
*/
int GfFile::checkNumCols(QFile *in)
{
/*
	looks for np
	
	rewinds the file back to the beginning
	
	tries to read the extended profile info first to get the file format
*/
char line[MAX_LINE_LEN];
float extended_version;

	while( ! in->atEnd() )
	{	
		in->readLine(line,MAX_LINE_LEN-1);
					
		if(line[0] == 'c' || line[0] =='C' )
		{			
			if(strstr(line,GFMAGIC) != NULL)
			{
				if(getfloatval(GFMAGIC,line,&extended_version) != GF_OK)
				{
					return(-1);
				}
							
				if(extended_version > GFEXTENDVER) // old version of code.
				{
					cout << "File format is newer than code, please upgrade\n";
					exit(-1);
				}

				in->readLine(line,MAX_LINE_LEN-1);
				if(in->atEnd())
				{
					return(-1);
				}
	
				if(gfprofile.readFormat(in,line)!= GF_OK)
				{
					return(-1);
				}
				else
				{
					break;
				}	

			}
		}
		else
			break;
	}

	// dump the comments
	
	while(line[0]== 'C' || line[0] == 'c')
	{
		in->readLine(line,MAX_LINE_LEN-1);
		if(in->atEnd())
		{
			return(-1);
		}			
	}
	
	//dump np
	
	switch(gfprofile.eof())
	{
	
	case EOF_AUTO:

       		switch(gfprofile.lineFormat())
        	{
	
			case FMT_FIXED:
				if (strlen(line) < 2*(unsigned int)gfprofile.dataLength() )  /* check for np */
           			{
					in->readLine(line,MAX_LINE_LEN-1);
					if(in->atEnd())
					{
						return(-1);
					}
				}	
				break;				
				
			
			case FMT_DELIM:
            			if(strchr(line,gfprofile.delimChar())==NULL)
            			{
					in->readLine(line,MAX_LINE_LEN-1);
					if(in->atEnd())
					{
						return(-1);
					}
				}
				break;
			
		}
	case EOF_NO:
		in->readLine(line,MAX_LINE_LEN-1);
		if(in->atEnd())
		{
			return(-1);
		}					
	break;
	case EOF_YES:
	break;
	}
	
	in->at(0); // back to beginning of the file
	
	return(chk_line_size(line));

}

int GfFile::chk_line_size(char *line)
{
int ncol,l;
char *dptr,*tline;
        if(gfprofile.lineFormat() == FMT_FIXED)
        {
             l=strlen(line);
             ncol = (int) ((l-gfprofile.invecLength())/((float)gfprofile.dataLength())+.5);
        }
        else
        {
             ncol = 1;
             tline=line;
strip2:
            if(tline[strlen(tline)-1] == gfprofile.delimChar() ) /* remove trailing delims */
            {
                tline[strlen(tline)-1] = '\0';
                goto strip2;
            }
more:
             dptr=strchr(tline,gfprofile.delimChar());
             if(dptr != NULL)
             {
                ncol++;
                tline=dptr+1;
strip:
                if(tline[0] == ' ')
                {
                        tline = tline +1;
                        goto strip;
                }
                goto more;
             }
	     
	     if(gfprofile.invecLength() > 0)
	     	ncol--;

        }

        return ncol;
}

	
/**
	Writes a gf file to filename.
<pre>
	Calls the QFile version.
	
	Returns:
		GF_BAD on i/o failure.
		GF_OK otherwise
</pre>
*/
int GfFile::write(char *filename)
{
int rtn;
	QFile qf(filename);
	
	if( qf.open(IO_WriteOnly ) )
	{
		rtn=write(&qf);
		qf.close();
		return(rtn);
	}
	else
		return(GF_BAD);
	
}

/**
	Writes a gf file to out.
<pre>
	Calls print with the stream pointed to the file.
	
	Returns:
		GF_BAD on i/o failure.
		GF_OK otherwise
</pre>
*/
int GfFile::write(QFile *out)
{
QTextStream ts(out);

	if(numPoints() > 0)
	{
		print(&ts);
		
		QFileInfo qfi(*out);
		file_name=qfi.fileName();

		return(GF_OK);
	}
	else
		return(GF_BAD);
}
	
/**	
	Replaces curve i (0 to numCurves()-1) with c.
<pre>
	Returns:
		GF_OK - success
		GF_BAD - i out of range.
</pre>
*/
int GfFile::setCurve(int i,const GfCurve &c)
{
	if( i < numcurves && i > -1)
	{
		*(curves[i]) = c;
		return(GF_OK);
	}
	else
		return(GF_BAD);
	
}

/**
	Adds c as the last GfCurve.
<pre>
	Always returnts GF_OK (out of memory will cause program to exit).
</pre>
*/
int GfFile::appendCurve(const GfCurve &c)
{
	storage(numcurves + 1);
	
	*(curves[numcurves - 1]) = c;
	
	return(GF_OK);
}

/**
	Adds c as the last GfCurve.
<pre>
	Always returnts GF_OK (out of memory will cause program to exit).
</pre>
*/
int GfFile::appendCurve(const GfCurve *c)
{
	storage(numcurves + 1);
	
	*(curves[numcurves - 1]) = (*c);
	
	return(GF_OK);
}

/**
	Removes the curve i (0 to numCurves()-1).
<pre>
	Returns:
		GF_OK -- success
		GF_BAD -- i out of range.
</pre>
*/
int GfFile::removeCurve(int i)
{
int j;
GfCurve *ctemp;

	if( i < numcurves && i > -1)
	{
		if( i < (numcurves - 1) )
		{
			ctemp = curves[i]; // save the one to be deleted	
	
			for(j=i;j<(numcurves-1);j++)
			{
				curves[j]=curves[j+1];
			}
			
			curves[numcurves-1] = ctemp; // put it last
		
		}
		
		storage(numcurves-1); // deletes the last curve
		
		return(GF_OK);
			
	}
	else
		return(GF_BAD);		
}

/**
	Breaks the curve c (0 to numCurves()-1) after point pt.
<pre>
	Returns:
		GF_OK -- success
		GF_BAD -- i out of range.
</pre>
*/
int GfFile::breakCurve(int c,int pt)
{
GfCurve ctemp;

	if( c < numcurves && c > -1)
	{
		if( (pt+1) < (curves[c]->numPoints()-1) )
		{
			ctemp.copySegment(curves[c],pt+1,curves[c]->numPoints()-1);
	
			appendCurve(ctemp);
		
			ctemp.copySegment(curves[c],0,pt);
			
			setCurve(c,ctemp);
		
			return(GF_OK);
		}
		else
			return(GF_BAD);
			
	}
	else
		return(GF_BAD);
}
/**
	Appends point (x,y) with invec value iv to curve c.
<pre>
	If c == numCurves a new curve is created.
	
	returns:
		GF_OK -- success
		GF_BAD -- c < 0 or c > numCurves()
</pre>
*/
int GfFile::appendPoint(int c,int iv,float x, float y)
{

	if(c == numcurves)
	{
		storage(numcurves + 1);
	
		curves[numcurves - 1]->appendPoint(iv,x,y);
		
		return(GF_OK);
	}
	
	if( c > -1 && c < numcurves)
	{
		curves[c]->appendPoint(iv,x,y);
		return(GF_OK);
	}
	
	return(GF_BAD);	
}

/**
	Appends a row of data to a curve with more than two columns of
	data storage.
<pre>
	c must be between 0 and numCurves()-1.  To append a curve with
	more than 2 columns, use appendCurve.
	
	No chacks are made that the array row has the proper amount of
	data.
	
	Returns:
		GF_OK
		GF_BAD -- c out of range.
</pre>
*/
int GfFile::appendRow(int c,int iv,float *row)
{
	/* this will not allocate a curve with the correct numcols
	if(c == numcurves)
	{
		storage(numcurves + 1);
	
		curves[numcurves - 1]->appendRow(iv,row);
		
		return(GF_OK);
	}
	*/
	if( c > -1 && c < numcurves)
	{
		curves[c]->appendRow(iv,row);
		return(GF_OK);
	}
	
	return(GF_BAD);	
}

/**
	Sets the columns to use for X and Y for curve c that has more than 
	two columns.
<pre>
	Default values of xcol, ycol are 1,2.
	
	Returns:
		GF_OK -- success
		GF_BAD -- xcol == ycol, xcol or ycol is too large or small.
</pre>
*/
int GfFile::setCurvePlotColumns(int c,int xcol,int ycol)
{
int ncol;

	if( c > -1 && c < numcurves)
	{
		ncol = curves[c]->numColumns();
		
		if(xcol < 0 || xcol >= ncol || ycol < 0 || ycol >= ncol 
			|| xcol == ycol)
			return(GF_BAD);
			
		if(curves[c]->setXColumn(xcol) == GF_BAD)
			return(GF_BAD);
			
		if(curves[c]->setYColumn(ycol) == GF_BAD)
			return(GF_BAD);
			
		return(GF_OK);
	}
	
	return(GF_BAD);

}
		
	
/**
	determines the closest data point on curve to the xm, ym screen coords.
<pre>
	Used for the edit selection routines.
	
	Returns:
		GF_BAD -- bad args or no point found
		GF_OK -- point found.
</pre>
*/
int GfFile::closestPoint(int xm,int ym,GfDraw *gfdraw,int *curve,int *point) const 
{
int xt,yt,i,j,pclose,cclose,dmin,dtest;

	if(numcurves < 1)
	{
		return(GF_BAD);
	}
	
	//xm = m->x();
	//ym = height-m->y();
	
	if(gfdraw->checkPixelCoords(xm,ym) == GF_NO)
	{
		return(GF_BAD);
	}
	
	dmin = 9999;
	pclose = 0;
	cclose = 0;

	for(j=0;j<numcurves;j++)
	{
		for(i=0;i<curves[j]->numPoints();i++)
		{
			xt = gfdraw->x_coord(curves[j]->pointX(i));
			yt = gfdraw->y_coord(curves[j]->pointY(i));
			
			dtest = abs(xm-xt)+abs(ym-yt);
			
			if(dtest < dmin )
			{
				dmin = dtest;
				pclose = i;
				cclose = j;
			}
		}
	}
	
	if(dmin < 9999)
	{
		*curve = cclose;
		*point = pclose;		
		return(GF_OK);
	}
	else
		return(GF_BAD);

}

/**
	Returns a GfCurve containing a segment of curve c from point start
	to point end.
<pre>
	Returns:
		GF_OK -- success (note: curve returned may have no points)
		GF_BAD -- on such curve.
</pre>
*/
GfCurve GfFile::curveSegment(int c,int start,int end) const 
{
GfCurve ans;

	if(c > -1 && c < numcurves)
	{		
		ans.copySegment(curves[c],start,end);
	}
	
	return(ans);
}

	
/**
	Sets the scale to s.
<pre>
	alwyas returns GF_OK.
</pre>
*/
int GfFile::setScale(const GfScale &s)
{
	xyscale = s;
	return(GF_OK);
}

/**
	Sets the internal scales to the max and min values of the data.
<pre>
	Also sets ticksType to TT_NONE, and linear scaling on both axes.
</pre>
*/
void GfFile::autoScale()
{
int i;
float xmax,xmin,ymax,ymin;
float xmaxc,xminc,ymaxc,yminc;

	if(numcurves == 0) return;
	
	xmax=1.;xmin =0.;ymax=1.;ymin=0.;
	
	for(i=0;i<numcurves;i++)
	{
		curves[i]->getMaxMin(&xmaxc,&xminc,&ymaxc,&yminc);
	
		if( i == 0)
		{
			xmax = xmaxc;
			ymax = ymaxc;
			xmin = xminc;
			ymin = yminc;
		}
		else
		{
			if(xmaxc > xmax) xmax = xmaxc;
			if(ymaxc > ymax) ymax = ymaxc;
			if(xminc < xmin) xmin = xminc;
			if(yminc < ymin) ymin = yminc;
		}
	
	}
	
	// these checks should really have a tolerance.
	
	if(xmax == xmin)
	{
		xmax = xmax + 0.5;
		xmin = xmin - 0.5;
	}
	
	if(ymax == ymin)
	{
		ymax = ymax + 0.5;
		ymin = ymin - 0.5;
	}
	
	xyscale.setXMaxMin(xmax,xmin);
	xyscale.setYMaxMin(ymax,ymin);
	
	xyscale.setXLogScale(GF_NO);
	xyscale.setYLogScale(GF_NO);
	
	xyscale.setTicksType(TT_NONE);

}
	
/**
	Sets the profile to p.
<pre>
	always returns GF_OK.
</pre>
*/
int GfFile::setProfile(const GfProfile &p)
{
	gfprofile = p;
	return(GF_OK);
}
	
/**
	Setsthe comments to c.
<pre>
	Always returns GF_OK.
</pre>
*/
int GfFile::setComments(const GfComments &c)
{
	gfcomments = c;
	return(GF_OK);
}

/**
	Sets the plotdefs to pd.
<pre>	
	Always returns GF_OK.
</pre>
*/
int GfFile::setPlotDefs(const GfPlotDefs &pd)
{
	gfplotdefs = pd;
	return(GF_OK);
}
	
/**
	Returns the total number of data points across all curves.
<pre>
</pre>
*/
int GfFile::numPoints() const 
{
int num,i;

	num = 0;

	for(i=0;i<numcurves;i++)
	{
		num = num + curves[i]->numPoints();
	}
	
	return(num);
}
	
/**
	Returns the number of GfCurves.
<pre>
</pre>
*/
int GfFile::numCurves() const 
{
	return(numcurves);
}

	
/**
	Returns the current GfScale.
<pre>
</pre>
*/
GfScale GfFile::scale() const 
{
	return(xyscale);
}
	
/**
	Returns the current GfProfile.
<pre>
</pre>
*/
GfProfile GfFile::profile() const 
{
	return(gfprofile);
}
	
/**
	Returns the current GfComments.
<pre>
</pre>
*/
GfComments GfFile::comments() const 
{
	return(gfcomments);
}

/**
	Returns the current GfPlotDefs.
<pre>
</pre>
*/
GfPlotDefs GfFile::plotDefs() const 
{
	return(gfplotdefs);
}
	
/**
	Returns curve number i, or an empty curve if i is out of range.
<pre>
</pre>
*/
const GfCurve &GfFile::curve(int i) const 
{

	if( i < numcurves && i > -1)
	{
		return(*(curves[i]));			
	}
	else
	{
		return(empty);
	}
}

/**
	Returns the lineType for curve i or GF_BAD if i is out of range.
<pre>
	Line types are defined in gfcommon.h.
</pre>
*/
int GfFile::curveLineType(int i) const 
{
	if( i < numcurves && i > -1)
	{
		return(curves[i]->lineType());			
	}
	else
	{
		return(GF_BAD);
	}
}

/**
	Returns the lineWidth of curve i or GF_BAD if out of range.
<pre>
	Line types are defined in gfcommon.h.
</pre>
*/
int GfFile::curveLineWidth(int i) const 
{
	if( i < numcurves && i > -1)
	{
		return(curves[i]->lineWidth());			
	}
	else
	{
		return(GF_BAD);
	}
}

/**
	Returns the color index of curve i or GF_BAD if i is out of range.
<pre>
	see gfcommon.h for color definitions.
</pre>
*/
int GfFile::curveColor(int i) const 
{
	if( i < numcurves && i > -1)
	{
		return(curves[i]->color());			
	}
	else
	{
		return(GF_BAD);
	}
}

/**
	Returns the line type pattern length for curve i or -1 on i out of range.
<pre>
	Pattern length is a float greater than 0.
</pre>
*/
float GfFile::curvePatternLength(int i) const 
{
	if( i < numcurves && i > -1)
	{
		return(curves[i]->patternLength());			
	}
	else
	{
		return(-1.);
	}
}

	
/**
	Sets the line type for curve i to t.
<pre>
	Returns:
		GF_OK -- success
		GF_BAD -- i out of range.
</pre>
*/
int GfFile::setCurveLineType(int i, int t)
{
	if( i < numcurves && i > -1)
	{
		curves[i]->setLineType(t);
		return(GF_OK);			
	}
	else
	{
		return(GF_BAD);
	}
}

/**
	Sets the line width for curve i to t.
<pre>
	Returns:
		GF_OK -- success
		GF_BAD -- i out of range.
</pre>
*/
int GfFile::setCurveLineWidth(int i, int t)
{
	if( i < numcurves && i > -1)
	{
		curves[i]->setLineWidth(t);
		return(GF_OK);			
	}
	else
	{
		return(GF_BAD);
	}
}

/**
	Sets the line color for curve i to t.
<pre>
	Returns:
		GF_OK -- success
		GF_BAD -- i out of range.
</pre>
*/
int GfFile::setCurveColor(int i, int t)
{
	if( i < numcurves && i > -1)
	{
		curves[i]->setColor(t);
		return(GF_OK);			
	}
	else
	{
		return(GF_BAD);
	}
}

/**
	Sets the line type pattern length for curve i to t.
<pre>
	Returns:
		GF_OK -- success
		GF_BAD -- i out of range.
</pre>
*/
int GfFile::setCurvePatternLength(int i, float t)
{
	if( i < numcurves && i > -1)
	{
		curves[i]->setPatternLength(t);
		return(GF_OK);			
	}
	else
	{
		return(GF_BAD);
	}
}

/**
	Returns the filename (if available) of the current Gf data file.
<pre>
</pre>
*/
QString GfFile::fileName() const 
{
	return(file_name);
}

/**
	Returns the comments from the header of the current Gf Data file.
<pre>
	This does not include the extended file format information.
</pre>
*/
QString GfFile::inFileComments() const 
{
	return(in_file_comments);
}

/**
	Sets the comments to be included at the top of the GF file.
<pre>	
	Each line of comments must begin with 'C' or 'c'.
</pre>
*/
void GfFile::setInFileComments(const char *text)
{
	in_file_comments.setStr(text);
}

/**
	Deletes all data and sets all variables to their default values.
*/
void GfFile::clear()
{
	storage(0);
	init();
}

void GfFile::test_mc()
{
int i;
float x[3];

GfCurve *c1;

	c1 = new GfCurve(3); // three columns GfCurve

	c1->setXColumn(0);
	c1->setYColumn(2); // lets see the sine curve

	for(i=0;i<100;i++)
	{
		x[0] = ((float)i)*3.14159/50.;
		x[1] = cos(x[0]);
		x[2] = sin(x[0]);
		
		c1->appendRow(1,x);
	}
	
	appendCurve(c1);
	
	autoScale();
	
}


void GfFile::test()
{
GfCurve c1;

	c1.appendPoint(0,1.,1.);
	c1.appendPoint(1,2.,1.);
	c1.appendPoint(1,2.,2.);
	c1.appendPoint(1,1.,2.);
	c1.appendPoint(1,1.,1.);
	
	appendCurve(c1);
	
GfCurve c2;

	c2.appendPoint(2,1.2,3.);
	c2.appendPoint(3,1.4,3.);
	c2.appendPoint(4,1.6,3.);
	c2.appendPoint(5,1.8,3.);
	c2.appendPoint(6,2.0,3.);
	c2.appendPoint(7,2.2,3.);
	c2.appendPoint(8,2.4,3.);
	c2.appendPoint(9,2.6,3.);
	c2.appendPoint(10,2.8,3.);
	c2.appendPoint(11,3.,3.);
	
	c2.setColor(3);
	
	appendCurve(c2);
		
GfScale s1;

	s1.setXMaxMin(3.,0.);
	s1.setYMaxMin(4.,0.);
	
	s1.setXLabel("x axis");
	s1.setYLabel("y axis");
	
	s1.setTicksType(TT_TICKS);
	s1.setXTicksStep(1.0);
	s1.setYTicksStep(0.5);

	setScale(s1);
	
GfComments com1;


	com1.addComment("test",20.,2.,LOCATION_ROWCOL);
	
	com1.addComment("2 x 2",2.,2.,LOCATION_DATA);
	
	setComments(com1);


	file_name.setStr("test");
}

/**
	Draws the Gf data to the GfDraw device gfd.
<pre>
	gfd must have been previously initialized to the desired type of
	output.
	
	It is up to the caller to execute gfd->done() when drawing is
	complete.
</pre>
*/
int GfFile::draw(GfDraw *gfd)
{
int i;

#ifdef DEBUG1
	cout << " drawing axes\n";
#endif	
	
	gfd->axes();
	
	
	for(i=0;i<numcurves;i++)
	{
#ifdef DEBUG1
	cout << "drawing curve: "<<i<<"\n";
#endif	
		gfd->curve(curves[i]);
	}
	
#ifdef DEBUG1
	cout << "drawing comments\n";
#endif	
	
	gfd->comments(&gfcomments);
	
#ifdef DEBUG1
	cout <<"done \n";
#endif	


	//gfd->done(); had to remove, because paint wants to draw selection
	// after this is called and done now deletes pdef.
	
	return(GF_OK);
	
}

/**
	Draws only a section of curve c from start to end with possibly a
	different color (offset by color_step) to gfd.
<pre>
	Does nothing if c or start/end are out of range.
</pre>
*/
void GfFile::drawSelection(GfDraw *gfd,int c,int start,int end,int color_step)
{
	if(c < 0 || c >=numcurves)
		return;
			
	gfd->curve(curves[c],start,end,color_step);
	
}

/**
	Returns the max and min data values in xmax,xmin,ymax,ymin for curve c.
<pre>
	Returns:
		GF_OK -- success
		GF_BAD -- c out of range.
</pre>
*/
int GfFile::getMaxMin(int c,float *xmax,float *xmin,float *ymax,float *ymin) const 
{
	if(c < 0 || c >=numcurves)
		return(GF_BAD);
		
	curves[c]->getMaxMin(xmax,xmin,ymax,ymin);
	
	return(GF_OK);
}
